with Ada.Text_IO; use Ada.Text_IO;

package body SoE is

   --  the upperbound to the search range
   --+ (NOTE: since no external entity needs access
   --+  to Limit, we can place its declaration in
   --+  the body of the SoE unit, so that we can
   --+  change its value at will without this causing
   --+  the specification of SoE to recompile)
   Limit : constant Positive := 10;

   --  each instance of process Sieve gets an own Id
   Assigned_Id : Natural := 0;

   --+-------------------------------------
   function New_Sieve return Sieve_Ref is
   begin
      Assigned_Id := Assigned_Id + 1;
      --  this function returns a new process instance
      --+ of type Sieve_T
      return new Sieve_T (Assigned_Id);
   end New_Sieve;
   --+-------------------------------------

   --+-------------------------------------
   task body Odd is
      --  process Odd has a child process S (instance of Sieve_T)
      --+ with which it will rendezvous for passing down the numbers
      --+ to search for primes in
      S : Sieve_Ref := New_Sieve;
      --  the first known prime number
      Num : Positive  := 3;
   begin
      --  process S is now activated and starts running
      --+ concurrently to Odd
      while Num < Limit loop
         --  process Odd passes all odd numbers found
         --+ in the range 3 .. Limit down to S
         --+ by means of rendezvous
         --+ (the 1st number passed down the line
         --+  is known to be a prime)
         S.Relay (Num);
         --  Odd slows down to rest a little in between
         delay 0.2;
         Num := Num + 2;
      end loop;
      --  process Odd completes but cannot terminate until
      --+ its child and dependent process S has terminated too
   end Odd;
   --+-----------------------

   --+-----------------------
   task body Sieve_T is
      Next_Sieve : Sieve_Ref;
      Prime,
      Num        : Positive;
   begin
      --  the process instance of Sieve_T accepts the 1st rendezvous
      --+ with its master, knowing that the 1st incoming value
      --+ will be a prime number
      accept Relay (Int : Positive) do
         Prime := Int;
      end Relay;
      Put_Line ("Sieve instance " & Natural'Image (Id) &
         " found prime number" & Natural'Image (Prime));
      loop
         --  the Sieve instance continues to accept rendezvous
         --+ with its master, to receive the other odd numbers
         --+ in the range
         accept Relay (Int : Positive) do
            Num := Int;
         end Relay;
         exit when Num rem Prime /= 0;
      end loop;
      --  any number which is not divisible by Prime
      --+ is itself a new prime, which this process instance
      --+ must pass down, by rendezvous, to a new child process,
      --+ instance of Sieve_T
      Next_Sieve := New_Sieve;
      Next_Sieve.Relay (Num);
      loop
         --  at this point this processs instace can accept
         --+ from its master all the remaining odd numbers
         --+ in the range and pass down to its child Sieve
         --+ those that are candidate primes because they are
         --+ not divisible by Prime
         accept Relay (Int : Positive) do
            Num := Int;
         end Relay;
         if Num rem Prime /= 0 then
            Next_Sieve.Relay (Num);
         end if;
      end loop;
      --  this process instance cannot complete because
      --+ it will attempt to rendezvous with its master without
      --+ knowing that it may in fact already have completed
      --+ (as Odd will certainly complete after having passed
      --+ down the last odd number in the range, its Sieve_T
      --+ child will never complete, so no other dependents will!)
   end Sieve_T;

end SoE;
